/* sexec.c -- execute compiled form of stream editor commands

   The single entry point of this module is the function execute(). It
may take a string argument (the name of a file to be used as text)  or
the argument NULL which tells it to filter standard input. It executes
the compiled commands in cmds[] on each line in turn.
   The function command() does most of the work. Match() and advance()
are used for matching text against precompiled regular expressions and
dosub() does right-hand-side substitution.  Getline() does text input;
readout() and memcmp() are output and string-comparison utilities.

====   Written for the GNU operating system by Eric S. Raymond  ====
====               For Kanji use by pcs19614 serow              ====
*/
/*
char id_sexec[] = "$Header: rcs/sexec.c 1.3 88/03/02 06:01:52 serow Exp $";
*/
/*	change any source program by LOGOS
*
* 1989.02/21 : command(), YCMD  when match break
* 1989.07/14 : match(), substitute() match times
* 1990.02/16 : memcmp() -> memcmps(), misc.
* 1990.12/15 : misc. correspond serow 88/03/02 & CR-X 90/09/14
* 1990.06/07 : execute(), check user interrupt
*/

#define __NO_STDIO_INLINE__
#include <stdio.h>	/* {f}puts, {f}printf, getk/putk, f{re}open, fclose */
#include <stdlib.h>
#include "kanji.h"      /* for kanji use */
#include "sed.h"	/* command type structures & miscellaneous constants */

/***** shared variables imported from the main ******/

/* main data areas */
extern khar	linebuf[];	/* current-line buffer */
extern sedcmd	*cmds; 	/* hold compiled commands */
extern long	linenum[];	/* numeric-addresses table */

/* miscellaneous shared variables */
extern int	nflag;		/* -n option flag */
extern int	eargc;		/* scratch copy of argument count */
extern sedcmd	*pending;	/* ptr to command waiting to be executed */

#if	GREP
extern	char grepcmd[];
extern	char grepmode;
#endif

/***** end of imported stuff *****/

/* functions */
/*
@IWił #pragma function(memcmp) ƂȂĂ邪T|[g
@ĂȂRpĈ memcmp  memcmps Ƃ
#pragma function(memcmp)
*/
static int memcmps();
static int selected();
static int match();
static int advance();


#define MAXHOLD 	MAXBUF	/* size of the hold space */
#define GENSIZ		MAXBUF	/* maximum genbuf size */

#define TRUE		1
#define FALSE		0

#ifndef BREAK
#define BREAK 1
#endif

#ifndef FLIST
#define FLIST 1
#endif

static char LTLMSG[]	= "sed: line too long\n";

static khar	*spend; 	/* current end-of-line-buffer pointer */
static long	lnum = 0L;	/* current source line number */
static long	llnum = 0L;	/* current source line number local */

/* append buffer maintenance */
static sedcmd	*appends[MAXAPPENDS];	/* array of ptrs to a,i,c commands */
static sedcmd	**aptr = appends;	/* ptr to current append */

/* genbuf and its pointers */
static khar	genbuf[GENSIZ];
static khar	*loc1;
static khar	*loc2;
static khar	*locs;

/* command-logic flags */
static int	lastline;		/* do-line flag */
static int	jump;			/* jump to cmd's link address if set */
static int	delete; 		/* delete command flag */

/* tagged-pattern tracking */
static khar	*bracend[MAXTAGS];	/* tagged pattern start pointers */
static khar	*brastart[MAXTAGS];	/* tagged pattern end pointers */

static char	ifname[260];		/* Input file name saved	*/

void
execute(file)		/* execute the compiled commands in cmds[] on a file */
char *file;		/* name of text source file to be filtered */
{
	register khar		*p1;		/* dummy copy ptrs */
	register sedcmd 	*ipc;		/* ptr to current command */
	khar			*execp; 	/* ptr to source */
	khar			*getline();	/* input-getting functions */
	void			command(), readout();
	char *strcpy();

	strcpy( ifname,(file!=NULL)?file:"(stdin)" );
#if FLIST != 0
	if (file != NULL){	/* filter text from a named file */
		if (freopen(file, "r", stdin) == NULL){
			exit(1);
		}
	}
#endif
	if (pending){		/* there's a command waiting */
		ipc = pending;		/* it will be first executed */
		pending = FALSE;	/*   turn off the waiting flag */
		goto doit;		/*   go to execute it immediately */
	}
	/* here's the main command-execution loop */
	llnum = 0L;			/* Clear local line number */
	for(;;){
		/* get next line to filter */
		if ((execp = getline(linebuf)) == BAD)
			return;
		spend = execp;
		/* loop through compiled commands, executing them */
		for(ipc = cmds; ipc->command; ){

#if BREAK
extern void test_signal(void);

			test_signal();			/* check ctrl-C */
#endif

			/* all no-address commands are selected */
			if (ipc->addr1 && !selected(ipc)){
				ipc++;
				continue;
			}
	doit:
			command(ipc);	/* execute the command pointed at */
			if (delete)	/* if delete flag is set */
				break;	/* don't exec rest of compiled cmds */
			if (jump){	/* if jump set, follow cmd's link */
				jump = FALSE;
				if ((ipc = ipc->u.link) == 0){
					ipc = cmds;
					break;
				}
			}else		/* normal goto next command */
				ipc++;
		}
		/* we've now done all modification commands on the line */
		/* here's where the transformed line is output */
		if (!nflag && !delete){
			for(p1 = linebuf; p1 < spend; p1++)
				putk(*p1,stdout);
			putk('\n',stdout);
		}
		/* if we've been set up for append, emit the text from it */
		if (aptr > appends)
			readout();
		delete = FALSE; /* clear delete flag; about to get next cmd */
	}
}

static int
selected(ipc)				/* is current command selected */
sedcmd	*ipc;
{
	register khar	*p1 = ipc->addr1;	/* point p1 at first address */
	register khar	*p2 = ipc->addr2;	/*   and p2 at second */
	khar		c;

	if (ipc->flags.inrange){
		if (*p2 == CEND){
			p1 = NULL;
		}else if (*p2 == CLNUM){
			c = p2[1];
			if (lnum > linenum[c]){
				ipc->flags.inrange = FALSE;
				if (ipc->flags.allbut)
					return(TRUE);
				ipc++;
				return(FALSE);
			}
			if (lnum == linenum[c])
				ipc->flags.inrange = FALSE;
		}else if (match(p2, 0))
			ipc->flags.inrange = FALSE;
	}else if (*p1 == CEND){
		if (!lastline){
			if (ipc->flags.allbut)
				return(TRUE);
			ipc++;
			return(FALSE);
		}
	}else if (*p1 == CLNUM){
		c = p1[1];
		if (lnum != linenum[c]){
			if (ipc->flags.allbut)
				return(TRUE);
			ipc++;
			return(FALSE);
		}
		if (p2)
			ipc->flags.inrange = TRUE;
	}else if (match(p1, 0)){
		if (p2)
			ipc->flags.inrange = TRUE;
	}else{
		if (ipc->flags.allbut)
			return(TRUE);
		ipc++;
		return(FALSE);
	}
	if (ipc->flags.allbut)
		return(FALSE);
	ipc++;
	return(TRUE);
}

static int
match(expbuf, gf)	/* uses genbuf */
/* match RE at expbuf against linebuf; if gf set, copy linebuf from genbuf */
khar	*expbuf;
int	gf;
{
	register khar	*p1, *p2, c;

	switch (gf){
	case 1:
		if (*expbuf)
			return(FALSE);
		p1 = linebuf; p2 = genbuf;
		while (*p1++ = *p2++);
		locs = p1 = loc2;
		break;
	case 0:
		p1 = linebuf;
		locs = FALSE;
		break;
	default:
		if (*expbuf)
			return(FALSE);
		locs = p1 = loc2;
	}

	p2 = expbuf;
	if (*p2++){
		loc1 = p1;
		if (*p2 == CCHR && p2[1] != *p1) /* 1st khar is wrong */
			return(FALSE);		/*   so fail */
		return(advance(p1, p2));	/* else try to match rest */
	}

	/* quick check for 1st kharacter if it's literal */
	if (*p2 == CCHR){
		c = p2[1];		/* pull out kharacter to search for */
		do {
			if (*p1 != c)
				continue;	/* scan the source string */
			if (advance(p1, p2))	/* found it, match the rest */
				return(loc1 = p1, 1);
		} while(*p1++);
		return(FALSE);		/* didn't find that first khar */
	}

	/* else try for unanchored match of the pattern */
	do {
		if (advance(p1, p2))
			return(loc1 = p1, 1);
	} while(*p1++);

	/* if got here, didn't match either way */
	return(FALSE);
}

static int advance(lp, ep)
/* attempt to advance match pointer by one pattern element */
register khar	*lp;		/* source (linebuf) ptr */
register khar	*ep;		/* regular expression element ptr */
{
	khar		*curlp; 	/* save ptr for closures */
	khar		c;		/* scratch kharacter holder */
	khar		*bbeg,*eptemp;
	int		ct,negflag,found;

	for (;;)
		switch (*ep++){
		case CCHR:		/* literal kharacter */
			if (*ep++ == *lp++)	/* if khars are equal */
				continue;	/* matched */
			return(FALSE);		/* else return false */

		case CDOT:		/* anything but newline */
			if (*lp++)		/* first NUL is at EOL */
				continue;	/* keep going if didn't find */
			return(FALSE);		/* else return false */

		case CNL:		/* start-of-line */
		case CDOL:		/* end-of-line */
			if (*lp == 0)		/* found that first NUL? */
				continue;	/* yes, keep going */
			return(FALSE);		/* else return false */

		case CEOF:		/* end-of-address mark */
			loc2 = lp;		/* set second loc */
			return(TRUE);		/* return true */

		case CKLIST:		/* a closure */
			c = *lp++;
			found = negflag = 0;
			if (*ep == CKNEG){
				ep++;
				negflag = 1;
			}
			while( *ep != CKEND ){
				switch( *ep++ ){
				case CKHAR:
					if (c == *ep)
						found = 1;
					ep++;
					break;
				case CKRANGE:
					if ((ep[0]<=c)&&(c<=ep[1]))
						found = 1;
					ep += 2;
					break;
				}
			}
			ep++;
			if ((negflag ^ found)&&(c))
				continue;
			else
				return(FALSE);	/* else return false */

		case CBRA:		/* start of tagged pattern */
			brastart[*ep++] = lp;	/* mark it */
			continue;		/* and go */

		case CKET:		/* end of tagged pattern */
			bracend[*ep++] = lp;	/* mark it */
			continue;		/* and go */

		case CBACK:
			bbeg = brastart[*ep];
			ct = bracend[*ep++] - bbeg;

			if (memcmps(bbeg, lp, ct)){
				lp += ct;
				continue;
			}
			return(FALSE);

		case CBACK|STAR:
			bbeg = brastart[*ep];
			ct = bracend[*ep++] - bbeg;
			curlp = lp;
			while(memcmps(bbeg, lp, ct))
				lp += ct;

			while(lp >= curlp){
				if (advance(lp, ep))
					return(TRUE);
				lp -= ct;
			}
			return(FALSE);


		case CDOT|STAR: 	/* match .* */
			curlp = lp;		/* save closure start loc */
			while (*lp++);		/* match anything */
			goto star;		/* now look for followers */

		case CCHR|STAR: 	/* match <literal khar>* */
			curlp = lp;		/* save closure start loc */
			while (*lp++ == *ep);	/* match many of that khar */
			ep++;			/* to start of next element */
			goto star;		/* match it and followers */

		case CKLIST|STAR:		/* match [...]* */
			curlp = lp;		/* save closure start loc */
			eptemp = ep;
			do{
				c = *lp++;
				ep = eptemp;
				found = negflag = 0;
				if (*ep == CKNEG){
					ep++;
					negflag = 1;
				}
				while( *ep != CKEND ){
					switch( *ep++ ){
					case CKHAR:
						if (c == *ep)
							found = 1;
						ep++;
						break;
					case CKRANGE:
						if ((ep[0]<=c)&&(c<=ep[1]))
							found = 1;
						ep += 2;
						break;
					}
				}
			}while( (negflag ^ found)&&(c) );
			ep++;
			goto star;		/* match followers */

star:		/* the recursion part of a * or + match */
			if (--lp == curlp)	/* 0 matches */
				continue;

			if (*ep == CCHR){
				c = ep[1];
				do {
					if (*lp != c)
						continue;
					if (advance(lp, ep))
						return(TRUE);
				} while(lp-- > curlp);
				return(FALSE);
			}

			if (*ep == CBACK){
				c = *(brastart[ep[1]]);
				do {
					if (*lp != c)
						continue;
					if (advance(lp, ep))
						return(TRUE);
				} while	(lp-- > curlp);
				return(FALSE);
			}
	
			do {
				if (lp == locs)
					break;
				if (advance(lp, ep))
					return(TRUE);
			} while(lp-- > curlp);
			return(FALSE);

		default:
			fprintf(stderr, "sed: RE error, %o\n", *--ep);
			exit(1);
		}
}

static int
substitute(ipc)
/* perform s command */
register sedcmd	*ipc;			/* ptr to s command struct */
{
	void dosub();			/* for if we find a match */
	unsigned char i, gf;

	i = 1;
	gf = ipc->flags.global;
	if (match(ipc->u.lhs, 0)) {		/* if no match */
		if (gf <= 1)			/* subst. when 1st or global */
			dosub(ipc->rhs);	/* perform it once */
		if (gf == 1)			/* 1st match */
			return(TRUE);
	}
	else
		return(FALSE);			/* command fails */

	while(*loc2)				/* cycle through possibles */
		if (gf) {
			if (match(ipc->u.lhs, 2)) {	/* found another */
				if (gf == ++i){		/* match any times */
					dosub(ipc->rhs);/* so subst. */
					break;
				}
				else if (gf < i)       /* not found in times */
					return(FALSE);
			} else
				return(FALSE);
		} else {
			if (match(ipc->u.lhs, 1)){	/* found another */
				dosub(ipc->rhs);	/* so substitute */
			} else				/* otherwise, */
				break;			/* we're done */
		}
	return(TRUE);					/* we succeeded */
}

void
dosub(rhsbuf)		/* uses linebuf, genbuf, spend */
/* generate substituted right-hand side (of s command) */
khar	*rhsbuf;	/* where to put the result */
{
	register khar	*lp, *sp, *rp;
	int		c,locscc;
	khar		*place(),*cplace();
	char		* np;
	char		nbuf[16];
	unsigned int strlen();

	locscc = loc2 - loc1;
	/* copy linebuf to genbuf up to location  1 */
	lp = linebuf; sp = genbuf;
	while (lp < loc1) *sp++ = *lp++;

	for (rp = rhsbuf; c = *rp++; ){
		if (c == '&'){
			sp = place(sp, loc1, loc2);
			continue;
		}else if ((c&0xFF00)==0x0100){
			c &= 0377;
			if (c == 'l' || c == 'L'){
				sprintf(nbuf,"%ld",c=='l'?llnum:lnum);
				sp = cplace( sp,nbuf,nbuf+strlen(nbuf) );
				continue;
			}else if (c == 'f'){
				if ((np = jrindex( ifname,'\\'))
				  ||(np = jrindex( ifname,'/' ))
				  ||(np = jrindex( ifname,':' ))){
					np++;
				}else{
					np = ifname;
				}
				sp = cplace( sp,np,np+strlen(np) );
				continue;
			}else if (c=='F'){
				sp = cplace( sp,ifname,ifname+strlen(ifname) );
				continue;
			}else if (('1' <= c)&&(c < MAXTAGS+'1')){
				c -= '1';
				sp = place(sp, brastart[c], bracend[c]);
				continue;
			}
		}
		*sp++ = c ;
		if (sp >= genbuf + MAXBUF)
			fprintf(stderr, LTLMSG);
	}
	lp = loc2;
	loc2 = linebuf + (sp-genbuf);
	while (*sp++ = *lp++)
		if (sp >= genbuf + MAXBUF)
			fprintf(stderr, LTLMSG);
	lp = linebuf; sp = genbuf;
	while (*lp++ = *sp++);
	spend = lp-1;
	if (locscc == 0) ++loc2;
}

khar *
place(asp, al1, al2)		/* uses genbuf */
/* place khars at *al1...*(al1 - 1) at asp... in genbuf[] */
register khar	*asp, *al1, *al2;
{
	while (al1 < al2){
		*asp++ = *al1++;
		if (asp >= genbuf + MAXBUF)
			fprintf(stderr, LTLMSG);
	}
	return(asp);
}

khar *
cplace(asp, al1, al2)		/* uses genbuf */
register khar	*asp;
register unsigned char	*al1, *al2;
{
	while (al1 < al2){
		*asp++ = *al1++;
		if (asp >= genbuf + MAXBUF)
			fprintf(stderr, LTLMSG);
	}
	return(asp);
}

void
listto(p1, fp)
/* write a hex dump expansion of *p1... to fp */
register khar	*p1;		/* the source */
FILE		*fp;		/* output stream to write to */
{
	p1--;
	while(*++p1){
		if (iskanji(*p1) || isprint(*p1)){
			putk(*p1, fp);		/* pass it through */
		}else{
			putk('\134', fp);		/* emit a backslash */
			switch(*p1){
			case '\10':	putk('b', fp); break;	/* BS */
			case '\11':	putk('t', fp); break;	/* TAB */
			case '\12':	putk('n', fp); break;	/* NL */
			case '\15':	putk('r', fp); break;	/* CR */
			case '\33':	putk('e', fp); break;	/* ESC */
			default:	fprintf(fp, "%02x", *p1 & 0xFF);
			}
		}
	}
	putk('\n', fp);
}

void
command(ipc)
/* execute compiled command pointed at by ipc */
sedcmd	*ipc;
{
	static int	didsub; 		/* true if last s succeeded */
	static khar	holdsp[MAXHOLD];	/* the hold space */
	static khar	*hspend = holdsp;	/* hold space end pointer */
	register khar	*p1, *p2;
	khar		*execp,*getline();
	void		readout();

	switch(ipc->command){
	case ACMD:		/* append */
		*aptr++ = ipc;
		if (aptr >= appends + MAXAPPENDS)
			fprintf(stderr,
				"sed: too many appends after line %ld\n",
				lnum);
		*aptr = 0;
		break;

	case CCMD:		/* change pattern space */
		delete = TRUE;
		if (!ipc->flags.inrange || lastline)
			puts((char *)ipc->u.lhs);	/* Not a putks */
		break;

	case DCMD:		/* delete pattern space */
		delete++;
		break;

	case CDCMD:		/* delete a line in hold space */
		p1 = p2 = linebuf;
		while(*p1 != '\n')
			if (delete = (*p1++ == 0))
				return;
		p1++;
		while(*p2++ = *p1++) continue;
		spend = p2-1;
		jump++;
		break;

	case EQCMD:		/* show current line number */
		fprintf(stdout, "%ld\n", lnum);
		break;

	case GCMD:		/* copy hold space to pattern space */
		p1 = linebuf;	p2 = holdsp;	while(*p1++ = *p2++);
		spend = p1-1;
		break;

	case CGCMD:		/* append hold space to pattern space */
		*spend++ = '\n';
		p1 = spend;	p2 = holdsp;
		while(*p1++ = *p2++)
			if (p1 >= linebuf + MAXBUF)
				break;
		spend = p1-1;
		break;

	case HCMD:		/* copy pattern space to hold space */
		p1 = holdsp;	p2 = linebuf;	while(*p1++ = *p2++);
		hspend = p1-1;
		break;

	case CHCMD:		/* append pattern space to hold space */
		*hspend++ = '\n';
		p1 = hspend;	p2 = linebuf;
		while(*p1++ = *p2++)
			if (p1 >= holdsp + MAXBUF)
				break;
		hspend = p1-1;
		break;

	case ICMD:		/* insert text */
		puts((char *)ipc->u.lhs);	/* Not 'putks'	*/
		break;

	case BCMD:		/* branch to label */
		jump = TRUE;
		break;

	case LCMD:		/* list text */
		listto(linebuf, (ipc->fout != NULL)?ipc->fout:stdout); break;

	case NCMD:	/* read next line into pattern space */
		if (!nflag)
			putks(linebuf);	/* flush out the current line */
		if (aptr > appends)
			readout();	/* do pending a, r commands */
		if ((execp = getline(linebuf)) == BAD){
			pending = ipc;
			delete = TRUE;
			break;
		}
		spend = execp;
		break;

	case CNCMD:	/* append next line to pattern space */
		if (aptr > appends)
			readout();
		*spend++ = '\n';
		if ((execp = getline(spend)) == BAD){
			pending = ipc;
			delete = TRUE;
			break;
		}
		spend = execp;
		break;

	case PCMD:		/* print pattern space */
#if	GREP
		if(grepmode) {
			printf("%s:%d:",ifname,llnum);
		}
#endif
		putks(linebuf);
		break;

	case CPCMD:		/* print one line from pattern space */
cpcom:				/* so s command can jump here */
		for(p1 = linebuf; *p1 != '\n' && *p1 != '\0'; )
			putk(*p1++,stdout);
		putk('\n',stdout);
		break;

	case QCMD:		/* quit the stream editor */
		if (!nflag)
			putks(linebuf);	/* flush out the current line */
		if (aptr > appends)
			readout();	/* do any pending a and r commands */
		exit(0);

	case RCMD:		/* read a file into the stream */
		*aptr++ = ipc;
		if (aptr >= appends + MAXAPPENDS)
			fprintf(stderr,
				"sed: too many reads after line %ld\n",
				lnum);
		*aptr = 0;
		break;

	case SCMD:		/* substitute RE */
		didsub = substitute(ipc);
		if (ipc->flags.print && didsub)
			if (ipc->flags.print == TRUE)
				putks(linebuf);
			else
				goto cpcom;
		if (didsub && ipc->fout){
			fputks(linebuf,ipc->fout);
			putk('\n',ipc->fout);
		}
		break;

	case TCMD:		/* branch on last s successful */
	case CTCMD:		/* branch on last s failed */
		if (didsub == (ipc->command == CTCMD))
			break;		/* no branch if last s failed, else */
		didsub = FALSE;
		jump = TRUE;		/*  set up to jump to assoc'd label */
		break;

	case CWCMD:		/* write one line from pattern space */
		for(p1 = linebuf; *p1 != '\n' && *p1 != '\0'; )
			putk(*p1++, ipc->fout);
		putk('\n', ipc->fout);
		break;

	case WCMD:		/* write pattern space to file */
		fputks(linebuf,ipc->fout);
		putk('\n',ipc->fout);
		break;

	case XCMD:		/* exchange pattern and hold spaces */
		p1 = linebuf;	p2 = genbuf;	while(*p2++ = *p1++) continue;
		p1 = holdsp;	p2 = linebuf;	while(*p2++ = *p1++) continue;
		spend = p2 - 1;
		p1 = genbuf;	p2 = holdsp;	while(*p2++ = *p1++) continue;
		hspend = p2 - 1;
		break;

	case YCMD:
		p1 = linebuf-1;
		while( *++p1 ){
			p2 = ipc->u.lhs;
			while( *p2 != CYEND ){
				if (*p1 == *p2++){
					*p1 = *p2;
					break;
				}
				++p2;
			}
		}
		break;
	}
}

khar *
getline(buf)
/* get next line of text to be filtered */
register khar	*buf;		/* where to send the input */
{
	static khar	lastk = 0;;

	if ((signed short)lastk != EOF && getks(buf) != NULL){
		lastk = getkhar();
		if ((signed short)lastk != EOF){
			ungetk(lastk);
		}else if (eargc==0){		/* Nothing folows?	*/
			lastline = TRUE;	/* This line is last one */
		}
		lnum++; 		/* note that we got another line */
		llnum++; 		/* note that we got another line */
		while(*buf++);		/* find the end of the input */
		return(--buf);		/* return ptr to terminating null */
	}else{
		if (eargc == 0) 		/* if no more args */
			lastline = TRUE;	/*    set a flag */
		lastk = 0;
		return(BAD);
	}
}

static int memcmps(a, b, count)
/* return TRUE if *a... == *b... for count khars, FALSE otherwise */
register khar	*a, *b;
register int	count;
{
	while(count--)			/* look at count kharacters */
		if (*a++ != *b++)	/* if any are nonequal	 */
			return(FALSE);	/*    return FALSE for false */
	return(TRUE);			/* compare succeeded */
}

void
readout()
/* write file indicated by r command to output */
{
	register int	t;	/* hold input khar or EOF */
	FILE		*fi;	/* ptr to file to be read */

	aptr = appends - 1;	/* arrange for pre-increment to work right */
	while(*++aptr){
		if ((*aptr)->command == ACMD){		/* process "a" cmd */
			puts((char *)(*aptr)->u.lhs);	/* Not a 'putks'   */
		}else{					/* process "r" cmd */
			if ((fi = fopen((char *)(*aptr)->u.lhs, "r")) == NULL)
				continue;
			while((t = getc(fi)) != EOF)	/* copy	, not 'getk'*/
				putc(t,stdout);	/* nor 'putk'	*/
			fclose(fi);
		}
	}
	aptr = appends; 	/* reset the append ptr */
	*aptr = 0;
}
